package com.fossgalaxy.games.tbs.metrics;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.PrintStream;
import java.time.Duration;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.UUID;
import java.util.function.Function;
import java.util.function.Predicate;

import com.fossgalaxy.games.tbs.App;
import com.fossgalaxy.games.tbs.GameDef;
import com.fossgalaxy.games.tbs.GameState;
import com.fossgalaxy.games.tbs.ai.AIFactory;
import com.fossgalaxy.games.tbs.ai.Controller;
import com.fossgalaxy.games.tbs.io.SettingsIO;
import com.fossgalaxy.games.tbs.io.map2.MapDef;
import com.fossgalaxy.games.tbs.metrics.parameters.EntityCost;
import com.fossgalaxy.games.tbs.metrics.parameters.EntityProp;
import com.fossgalaxy.games.tbs.metrics.parameters.ParameterList;
import com.fossgalaxy.games.tbs.metrics.parameters.Parameter;
import com.fossgalaxy.games.tbs.order.Order;
import com.fossgalaxy.games.tbs.order.OrderProcessor;
import com.fossgalaxy.games.tbs.parameters.GameSettings;
import com.fossgalaxy.games.tbs.rules.Rule;

import io.jenetics.IntegerGene;
import io.jenetics.engine.Codec;
import io.jenetics.engine.Engine;
import io.jenetics.engine.EvolutionResult;
import io.jenetics.engine.EvolutionStream;
import io.jenetics.engine.Limits;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public class AppEvolver {
	public static Logger LOG = LoggerFactory.getLogger(AppEvolver.class);

	protected final ParameterList list;
	protected final GameSettings common;
	protected final SettingsIO io;
	
	protected AIFactory ai;
	protected MapDef map;
	
	protected Map<GameSettings, Double> fitnessScores;
	
	public static void main(String[] args) {
		
		SettingsIO io = new SettingsIO();
		GameDef def = io.loadGameDef("game.json");
		AIFactory ai = new AIFactory(io, def.getEvalFileName(), def.getRuleFileName(), def.getAiFileName());
		
		//we need to know the settings from the file
		GameSettings settings = io.loadSettings(def);
		
		//build the evolver using our base game
		AppEvolver evo = new AppEvolver(io, settings, ai, "map_balance.json");
		
		//tweak the red knight's defence between 0 and 10 in steps of 1
		evo.addParameter(new EntityProp("red_knight", "defRanged", 0, 10, 1)); 
		
		//tweak the blue archer's attack between 0 and 10 in steps of 2
		evo.addParameter(new EntityProp("blue_archer", "atkRanged", 0, 10, 1));
		
		GameSettings best = evo.evolve(evo::evaluateWithFiles);
		
		//print out the settings the GA picked
		System.out.println(best.getEntityType("blue_archer").getProperty("atkRanged"));
		System.out.println(best.getEntityType("red_knight").getProperty("defRanged"));
		
	}
	
	public AppEvolver(SettingsIO io, GameSettings common, AIFactory ai, String mapFile) {
		this.list = new ParameterList();
		this.fitnessScores = new HashMap<>();
		
		this.io = io;
		this.common = common;
		
		this.ai = ai;
		
		this.map = io.loadMapDef(mapFile);
	}
	
	public void addParameter(Parameter p ) {
		list.add(p);
	}
	
	public Map<GameSettings, Double> getAllFitnesses(){
		return fitnessScores;
	}
	
	//if you want to tweak the GA, you can override this function
	public Engine<IntegerGene, Double> buildEngine(Function<GameSettings, Double> eval, Codec<GameSettings, IntegerGene> codec) {
		return Engine.builder(eval, codec)
				.populationSize(20)
				.maximizing()
				.build();
	}
	
	//if you want to tweak time limits, etc... override this.
	public List<Predicate<? super EvolutionResult<IntegerGene, Double>>> getLimits() {
		return Arrays.asList(
				Limits.byFixedGeneration(10),
				Limits.byExecutionTime(Duration.ofMinutes(5))
				);
	}
	
	//the actual evolution function, you shouldn't need to touch this.
	public GameSettings evolve(Function<GameSettings, Double> eval) {
		
		Codec<GameSettings, IntegerGene> codec = new GeneDecoder(list, common, io);
		
		Engine<IntegerGene, Double> engine = buildEngine(eval, codec);
		
		EvolutionStream<IntegerGene, Double> stream = engine.stream();
		for (Predicate<? super EvolutionResult<IntegerGene, Double>> limit : getLimits() ) {
			stream = stream.limit(limit);
		}
		
		return stream.collect(EvolutionResult.toBestResult(codec));
	}
	
	/**
	 * How good is this set of parameters?
	 * 
	 * @param settings the parameters we're considering
	 * @return the fitness for this game
	 */
	public Double evaluate(GameSettings settings) {
		
		GameState start = map.buildState(settings);
		
		int[] winCounts = new int[2];
		
		for (int i=0; i<10; i++) {
			
			Controller[] controllers = new Controller[] {
					ai.buildAI("RangedRush", settings),
					ai.buildAI("KnightRush", settings)
			};
			
			GameState state = new GameState(start);
			GameMetrics metrics = runGame(state, settings, controllers);
		
			Integer winner = metrics.getWinner();
			if (winner != null) {
				winCounts[winner]++;
			}
			
		}
		
		double score = winCounts[1] - winCounts[0];
		fitnessScores.put(settings, score);
		
		return score;
	}
	
	public Double evaluateWithFiles(GameSettings settings) {
		
		GameState start = map.buildState(settings);
		
		PrintStream out = null;
		try {
			File logFile = new File(String.format("logs/%s.csv", UUID.randomUUID()));
			out = new PrintStream(logFile);
		} catch (FileNotFoundException ex) {
			System.err.println("You didn't create the logs/ directory!");
			System.exit(1);
		}
		
		int defence = settings.getEntityType("red_knight").getProperty("defRanged");
		int attack = settings.getEntityType("blue_archer").getProperty("atkRanged");
		
		
		int[] winCounts = new int[2];
		int[] peiceDifference = new int[2];
		
		for (int i=0; i<10; i++) {
			
			Controller[] controllers = new Controller[] {
					ai.buildAI("RangedRush", settings),
					ai.buildAI("KnightRush", settings)
			};
			
			GameState state = new GameState(start);
			GameMetrics metrics = runGame(state, settings, controllers);
		
			Integer winner = metrics.getWinner();
			if (winner != null) {
				winCounts[winner]++;
			}
			
			//an example of more advanced metric - peice difference
			List<String> p1Entities = new ArrayList<>();
			List<String> p2Entities = new ArrayList<>();
			
			for (int turn=0; turn<metrics.getTurns(); turn++) {
				TurnMetrics turnMetrics = metrics.getMetricsForTurn(turn);
				
				GameState startState = turnMetrics.getStateAtStart();
				GameState endState = turnMetrics.getStateAtEnd();
				
				p1Entities.add( Integer.toString(endState.getOwnedEntities(0).size()) );
				p2Entities.add( Integer.toString(endState.getOwnedEntities(1).size()) );
				
				int peices = endState.getEntities().size() -  startState.getEntities().size();
				
				peiceDifference[turnMetrics.getPlayerID()] += peices;
			}
			
			if (out != null) {
				out.printf("%d,%d,%d,%d,%s%n", 0, i, attack, defence, String.join(",", p1Entities));
				out.printf("%d,%d,%d,%d,%s%n", 1, i, attack, defence, String.join(",", p2Entities));
			}
				
		}
		
		if (out != null) {
			out.close();
		}
		
		double score = winCounts[1] - winCounts[0];
		fitnessScores.put(settings, score);
		
		return score;
	}
	
    public static GameMetrics runGame(GameState level, GameSettings settings, Controller[] controllers) {
    	OrderProcessor processor = new OrderProcessor(level, controllers.length);
    	
    	for (Rule rule : settings.getVictoryConditions()) {
    		processor.addRule(rule);
    	}
    	
    	for (int i=0; i < settings.getTurnLimit(); i++) {
    		processor.setupTurn();
    		
    		int playerID = processor.getCurrentPlayer();
    		Controller controller = controllers[playerID];
    		
    		GameState stateClone = new GameState(level);
    		
    		Map<UUID, Order> orders = controller.doTurn(playerID, stateClone);
    		processor.doOrderBulk(orders);

    		processor.finishTurn();
    		
    		if (!Objects.equals(processor.getWinner(), Rule.NO_WINNER)) {
    			break;
    		}
    	}
    	
    	return processor.getMetrics();
    }

	public static void dumpToFile(AppEvolver evo, String filename) {
		try {

			File file = new File(filename);

			//i'd rather you didn't lose the data - i'll make sure the directory exists...
			File dir = new File(file.getParent());
			if (!dir.exists()) {
				if (!dir.mkdir()) {
					System.err.println("directory didn't exist and I couldn't make it :(");
				}
			}

			PrintStream ps = new PrintStream(file);

			List<String> headings = new ArrayList<>();
			for (Parameter p : evo.getParameters()) {
				headings.add("\"" + p.getPrettyName() + "\"");
			}
			headings.add("fitness");
			ps.println(String.join(",", headings));

			for (Map.Entry<GameSettings, Double> fitnessEntry : evo.getAllFitnesses().entrySet()) {
				GameSettings settings = fitnessEntry.getKey();

				List<String> parameters = new ArrayList<>();
				for (Parameter p : evo.getParameters()) {
					parameters.add(Integer.toString(p.extractValue(settings)));
				}
				parameters.add(Double.toString(fitnessEntry.getValue()));
				ps.println(String.join(",", parameters));
			}

			ps.close();
		} catch (FileNotFoundException ex) {
			ex.printStackTrace();
		}
	}

	public ParameterList getParameters() {
		return list;
	}
}
